using Microsoft.Extensions.Configuration;
using System.Text.Json;

namespace View_by_Distance.Duplicate.Search.Models.Binder;

public class AppSettings
{

#nullable disable

    public string Company { get; set; }
    public bool? IndexOnly { get; set; }
    public string[] IgnoreRelativePaths { get; set; }
    public int? MaxDegreeOfParallelism { get; set; }
    public string OutputExtension { get; set; }
    public bool? SortContainers { get; set; }
    public bool? Reverse { get; set; }
    public string WorkingDirectoryName { get; set; }

#nullable restore

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

    private static Models.AppSettings Get(AppSettings? appSettings)
    {
        Models.AppSettings result;
        if (appSettings?.IndexOnly is null)
            throw new NullReferenceException(nameof(appSettings.IndexOnly));
        if (appSettings?.IgnoreRelativePaths is null)
            throw new NullReferenceException(nameof(appSettings.IgnoreRelativePaths));
        if (appSettings?.MaxDegreeOfParallelism is null)
            throw new NullReferenceException(nameof(appSettings.MaxDegreeOfParallelism));
        if (appSettings?.SortContainers is null)
            throw new NullReferenceException(nameof(appSettings.SortContainers));
        if (appSettings?.Reverse is null)
            throw new NullReferenceException(nameof(appSettings.Reverse));
        result = new(
            appSettings.Company,
            appSettings.IndexOnly.Value,
            appSettings.IgnoreRelativePaths,
            appSettings.MaxDegreeOfParallelism.Value,
            appSettings.OutputExtension,
            appSettings.SortContainers.Value,
            appSettings.Reverse.Value,
            appSettings.WorkingDirectoryName
        );
        return result;
    }

    public static Models.AppSettings Get(IConfigurationRoot configurationRoot)
    {
        Models.AppSettings result;
        AppSettings? appSettings = configurationRoot.Get<AppSettings>();
        result = Get(appSettings);
        return result;
    }

}