using System.Text.Json;
using System.Text.Json.Serialization;

namespace View_by_Distance.Shared.Models;

public class PersonContainer : Properties.IPersonContainer
{

    public int? ApproximateYears { init; get; }
    public Person? Person { init; get; }
    public PersonBirthday[]? Birthdays { init; get; }
    public string[] DisplayDirectoryAllFiles { init; get; }
    public string DisplayDirectoryName { init; get; }
    public long? Key { init; get; }
    public char? Char { init; get; }

    [JsonConstructor]
    public PersonContainer(int? approximateYears, char? @char, Person? person, PersonBirthday[]? birthdays, string[] displayDirectoryAllFiles, string displayDirectoryName, long? key)
    {
        ApproximateYears = approximateYears;
        Char = @char;
        Person = person;
        Birthdays = birthdays;
        DisplayDirectoryAllFiles = displayDirectoryAllFiles;
        DisplayDirectoryName = displayDirectoryName;
        Key = key;
    }

    public PersonContainer(char[] personCharacters, PersonBirthday birthday, string displayDirectoryName) :
        this(Stateless.Methods.IAge.GetApproximateYears(personCharacters, displayDirectoryName), null, Stateless.Methods.IPerson.GetPerson(personCharacters, displayDirectoryName, birthday.Value.Ticks, birthday), new PersonBirthday[] { birthday }, Array.Empty<string>(), displayDirectoryName, birthday.Value.Ticks)
    { }

    public PersonContainer(int? approximateYears, PersonBirthday birthdays, string displayDirectoryName, long key) :
        this(approximateYears, null, null, new PersonBirthday[] { birthdays }, Array.Empty<string>(), displayDirectoryName, key)
    { }

    public PersonContainer(int? approximateYears, PersonBirthday birthdays, char? @char, string displayDirectoryName, long key) :
        this(approximateYears, @char, null, new PersonBirthday[] { birthdays }, Array.Empty<string>(), displayDirectoryName, key)
    { }

    public PersonContainer(int? approximateYears, char @char, string[] displayDirectoryAllFiles, string displayDirectoryName) :
        this(approximateYears, @char, null, null, displayDirectoryAllFiles, displayDirectoryName, null)
    { }

    public PersonContainer(int? approximateYears, Person? person, PersonBirthday[]? birthdays, string[] displayDirectoryAllFiles, string displayDirectoryName, long? key) :
        this(approximateYears, null, person, birthdays, displayDirectoryAllFiles, displayDirectoryName, key)
    { }

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

}